"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 7, 2013.

Package initialization file.
"""
from .repository import FileSystemRepository as Repository

